@@ -1,7 +1,7 @@
 /*
  * tcpd_local.h 
  *
- * 		$Id: tcpd_local.h,v 1.4 2000/05/15 12:38:55 misiek Exp $
+ * 		$Id: tcpd_local.h,v 1.6 2000/10/01 21:19:56 baggins Exp $
  *
  *		This program is free software; you can redistribute it and/or
  *		modify it under the terms of the BSD style license (see
@@ -33,6 +33,8 @@
 #define	T_FUNC_STR1	6  /* exec function with string option 1 */
 #define	T_FUNC_STR2	7  /* exec function with string option 2 */
 #define	T_FUNC_UINT1	8  /* exec function with unsigned integer */
+#define	T_FUNC_SET0	9  /* exec function with 0 as a parameter */
+#define	T_FUNC_SET1	10 /* exec function with 1 as a parameter */
 
 #ifndef SA_LEN
 #define SA_LEN(x)       ((x)->sa_len)
@@ -58,5 +60,6 @@ void prefixlen2mask(int prefixlen, unsigned char *dest, int size);
 int tcpd_parse_options(char *p, char **left, struct _code *flags);
 int tcpd_get_token2(char *str, char **left, char *buf, int buflen);
 int tcpd_parse_syslog(char *str, char **left);
+char *tcpd_lineno(struct request_info *req);
 
 #endif /* tcpd_local.h */
