@@ -14,7 +14,7 @@
 #endif
 
 #include "rcsid.h"
-RCSID(PKG_VER "$Id: libwrap.c,v 1.7 2000/05/15 09:26:13 misiek Exp $")
+RCSID(PKG_VER "$Id: libwrap.c,v 1.17 2000/10/01 21:19:57 baggins Exp $")
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -34,10 +34,22 @@ RCSID(PKG_VER "$Id: libwrap.c,v 1.7 2000/05/15 09:26:13 misiek Exp $")
 #include "tcpd.h"
 #include "tcpd_local.h"
 
+struct __service_options {
+	int refuse_wait;
+	int rfc1413_timeout;
+	int allow_severity;
+	int deny_severity;
+	int ident;
+	int defaultrule;
+	int ignore_src_dns_fake;
+	int ignore_dst_dns_fake;
+};
+
 static char access_table[TCPD_STRINGLENL] = "/etc/hosts.access";
 static int userresolvreq;
 static int defaultallow;
-static int ignore_dns_fake;
+static int ignore_src_dns_fake;
+static int ignore_dst_dns_fake;
 
 static void strtosa(char *str, struct sockaddr_storage *sa)
 {
@@ -85,6 +97,7 @@ static void mappedtov4(struct sockaddr_storage *ss)
 		
 static struct request_info *request_fill(struct request_info *req, va_list ap)
 {
+	struct sockaddr_storage *ss;
 	int cmd;
 
 	do {
@@ -100,8 +113,14 @@ static struct request_info *request_fill(struct request_info *req, va_list ap)
 			strncpy(req->user, va_arg(ap, void *), sizeof(req->user));
 			break;
 		case RQ_SERVER_SIN:
+			ss = va_arg(ap, struct sockaddr_storage *);
+			if (ss)
+				memcpy(&req->dst, ss, SA_LEN((struct sockaddr *)ss));
+			break;
 		case RQ_CLIENT_SIN:
-			(void)va_arg(ap, void *);
+			ss = va_arg(ap, struct sockaddr_storage *);
+			if (ss)
+				memcpy(&req->src, ss, SA_LEN((struct sockaddr *)ss));
 			break;
 		case RQ_SERVER_NAME:
 			strncpy(req->dst_name, va_arg(ap, void *), sizeof(req->dst_name));
@@ -137,7 +156,8 @@ struct request_info *request_init(struct request_info *req, ...)
 
 	userresolvreq = 1;
 	defaultallow = TCPD_DENY;
-	ignore_dns_fake = 0;
+	ignore_src_dns_fake = 0;
+	ignore_dst_dns_fake = 0;
 
 	memset(req, 0, sizeof(struct request_info));
 	req->sock = -1;
@@ -147,7 +167,8 @@ struct request_info *request_init(struct request_info *req, ...)
 	allow_severity = deny_severity = LOG_DAEMON | LOG_NOTICE;
 	req->udp = 0;
 	req->refuse_wait = 5;
-	req->wrongdns = 0;
+	req->src_wrongdns = 0;
+	req->dst_wrongdns = 0;
 	req->rfc1413_timeout = TCPD_DEFAULT_RFC1413_TIMEOUT;
 	r = request_fill(req, ap);
 	va_end(ap);
@@ -219,9 +240,9 @@ void fromhost(struct request_info *req)
 	mappedtov4(&req->dst);
 	mappedtov4(&req->src);
 	if (gethostaddr(&req->src, req->src_name, sizeof(req->src_name)) == -1)
-		req->wrongdns = 1;
+		req->src_wrongdns = 1;
 	if (gethostaddr(&req->dst, req->dst_name, sizeof(req->dst_name)) == -1)
-		req->wrongdns = 1;
+		req->dst_wrongdns = 1;
 }
 
 /*
@@ -308,8 +329,8 @@ static int match_host_part(struct request_info *req, char *user, char *host, int
 			return TCPD_MATCH;
 		if (*host == '.') {
 			p = req->src_name;
-			while((p = strchr(p, '.')) != NULL)
-				if (strcasecmp(p, host))
+			while((p = strchr(++p, '.')) != NULL)
+				if (!strcasecmp(p, host))
 					return TCPD_MATCH;
 			return TCPD_NOMATCH;
 		}
@@ -444,55 +465,67 @@ static int host_match(struct request_info *req, char *strp, char **left)
  */
 
 static struct request_info *sreq;
+static struct __service_options ALL_service = {-1, -1, -1, -1, -1, -1, -1};
+static struct __service_options current_service = {-1, -1, -1, -1, -1, -1, -1};
+static struct __service_options *sservicep = &ALL_service;
 
 static int set_syslogallow(char *s)
 {
-  int i = tcpd_parse_syslog(s, NULL);
+	int i = tcpd_parse_syslog(s, NULL);
 
-  if (i >= 0) {
-    allow_severity = i;
-    return 1;
-  }
-  return 0;
+	if (i >= 0) {
+		sservicep->allow_severity = i;
+		return 1;
+	}
+	return 0;
 }
 
 static int set_syslogdeny(char *s)
 {
-  int i = tcpd_parse_syslog(s, NULL);
+	int i = tcpd_parse_syslog(s, NULL);
 
-  if (i >= 0) {
-    deny_severity = i;
-    return 1;
-  }
-  return 0;
+	if (i >= 0) {
+		sservicep->deny_severity = i;
+		return 1;
+	}
+	return 0;
 }
 
 static int set_syslog(char *s)
 {
-  int i = tcpd_parse_syslog(s, NULL);
+	int i = tcpd_parse_syslog(s, NULL);
 
-  if (i >= 0) {
-    allow_severity = deny_severity = i;
-    return 1;
-  }
-  return 0;
+	if (i >= 0) {
+		sservicep->allow_severity = sservicep->deny_severity = i;
+		return 1;
+	}
+	return 0;
 }
 
-static int set_refuse_wait(int val){sreq->refuse_wait = val; return 1;}
-static int set_rfc1413_timeout(int val){sreq->rfc1413_timeout = val;return 1;}
+static int set_refuse_wait(int val) { sservicep->refuse_wait = val; return 1; }
+static int set_rfc1413_timeout(int val) { sservicep->rfc1413_timeout = val; return 1; }
+static int set_ident(int val) { sservicep->ident = val; return 1; }
+static int set_defaultrule(int val) { sservicep->defaultrule = val; return 1; }
+static int set_dns(int val) { sservicep->ignore_src_dns_fake = sservicep->ignore_dst_dns_fake = val; return 1; }
+static int set_srcdns(int val) { sservicep->ignore_src_dns_fake = val; return 1; }
+static int set_dstdns(int val) { sservicep->ignore_dst_dns_fake = val; return 1; }
 
 struct _code flagsheader[] = {
 	{ "syslogallow", T_FUNC_STR1, &set_syslogallow},
 	{ "syslogdeny", T_FUNC_STR1, &set_syslogdeny},
 	{ "syslog", T_FUNC_STR1, &set_syslog},
-	{ "defaultallow", T_SET1, &defaultallow },
-	{ "defaultdeny", T_SET0, &defaultallow },
+	{ "defaultallow", T_FUNC_SET1, &set_defaultrule },
+	{ "defaultdeny", T_FUNC_SET0, &set_defaultrule },
 	{ "refuse_wait", T_FUNC_UINT1, &set_refuse_wait },
 	{ "rfc1413_timeout", T_FUNC_UINT1, &set_rfc1413_timeout },
-	{ "ignorednsfake", T_SET1, &ignore_dns_fake },
-	{ "denydnsfake", T_SET0, &ignore_dns_fake },
-	{ "ident", T_SET1, &userresolvreq },
-	{ "noident", T_SET0, &userresolvreq },
+	{ "ignorednsfake", T_FUNC_SET1, &set_dns },
+	{ "ignoresrcdnsfake", T_FUNC_SET1, &set_srcdns },
+	{ "ignoredstdnsfake", T_FUNC_SET1, &set_dstdns },
+	{ "denydnsfake", T_FUNC_SET0, &set_dns },
+	{ "denysrcdnsfake", T_FUNC_SET0, &set_srcdns },
+	{ "denydstdnsfake", T_FUNC_SET0, &set_dstdns },
+	{ "ident", T_FUNC_SET1, &set_ident },
+	{ "noident", T_FUNC_SET0, &set_ident },
 	{ NULL }
 };
 
@@ -521,7 +554,6 @@ static int tcpd_parse_option(struct request_info *req, char *str, int mode)
 	return TCPD_MATCH;
 }
 
-
 static int hosts_access0(FILE *fp, struct request_info *req)
 {
 	char buff[TCPD_BUFLEN];
@@ -531,6 +563,7 @@ static int hosts_access0(FILE *fp, struct request_info *req)
 #define SERVICE_OTHER 0
 	int in_service = SERVICE_ALL;
 	char service_name[TCPD_STRINGLEN];
+	int result = defaultallow;
 
 	*service_name = '\0';
 	while(fgets(buff, sizeof(buff), fp) != NULL) {
@@ -551,7 +584,10 @@ static int hosts_access0(FILE *fp, struct request_info *req)
 					tcpd_to_nextseparator(&r);
 					if (tcpd_parse_option(req, r, 1) == TCPD_ERROR)
 						return TCPD_ERROR; /* parse error */
-					return TCPD_ALLOW;
+					tcpd_to_nextseparator(&r);
+					if (tcpd_exec_option(req) == TCPD_ERROR)
+						return TCPD_ERROR; /* parse error */
+					result = TCPD_ALLOW;
 				case TCPD_CONTINUE:
 					continue;
 				case TCPD_ERROR:
@@ -569,7 +605,10 @@ static int hosts_access0(FILE *fp, struct request_info *req)
 					tcpd_to_nextseparator(&r);
 					if (tcpd_parse_option(req, r, 1) == TCPD_ERROR)
 						return TCPD_ERROR; /* parse error */
-					return TCPD_DENY;
+					tcpd_to_nextseparator(&r);
+					if (tcpd_exec_option(req) == TCPD_ERROR)
+						return TCPD_ERROR; /* parse error */
+					result = TCPD_DENY;
 				case TCPD_CONTINUE:
 					continue;
 				case TCPD_ERROR:
@@ -585,8 +624,12 @@ static int hosts_access0(FILE *fp, struct request_info *req)
 				return TCPD_ERROR; /* parse error */
 			continue;
 		}
-		if (!tcpd_word(p, service = "ALL", &q) )
+		if (!tcpd_word(p, service = "ALL", &q) ) {
 			service = req->service;
+			sservicep = &current_service;
+		} else {
+			sservicep = &ALL_service;
+		}
 		switch(service_match(p, service, req, &q)) {
 		case TCPD_MATCH:
 			if (*q != '\0') {
@@ -595,25 +638,29 @@ static int hosts_access0(FILE *fp, struct request_info *req)
 				else
 					return TCPD_ERROR;
 			}
-			if (in_service == SERVICE_IN
-			    && strcasecmp(p, service_name))
-				return defaultallow;
-			in_service = SERVICE_IN;
-			strncpy(service_name, p, sizeof(service_name));
+			/* Next service? */
+			if (in_service != SERVICE_IN ||
+					!strcasecmp(p, service_name)) {
+				in_service = SERVICE_IN;
+				strncpy(service_name, p, sizeof(service_name));
+			}
 			SPACESKIP(q);
 			if (tcpd_parse_option(req, q, 0) == TCPD_ERROR)
 				return TCPD_ERROR; /* parse error */
 			continue;
 		case TCPD_NOMATCH:
-			if (in_service == SERVICE_IN)
-				return defaultallow;
+			/* Next service */
+			if (in_service == SERVICE_IN) {
+				in_service = SERVICE_OTHER;
+				continue;
+			}
 		default:
 			break;
 		}
 		in_service = SERVICE_OTHER;
 	}
 	req->lineno = -1;
-	return defaultallow;
+	return result;
 }
 
 int hosts_access(struct request_info *req)
@@ -631,7 +678,21 @@ int hosts_access(struct request_info *req)
 	ret = hosts_access0(fp, req);
 	fclose(fp);
 
-	if (ret == TCPD_ALLOW && !ignore_dns_fake && req->wrongdns)
+#define SET_VAR(g,a,c) ((((a)<0)&&((c)<0)) ? g : (((c)<0) ? a : c))
+
+	req->refuse_wait = SET_VAR(req->refuse_wait, ALL_service.refuse_wait, current_service.refuse_wait);
+	req->rfc1413_timeout = SET_VAR(req->rfc1413_timeout, ALL_service.rfc1413_timeout, current_service.rfc1413_timeout);
+	allow_severity = SET_VAR(allow_severity, ALL_service.allow_severity, current_service.allow_severity);
+	deny_severity = SET_VAR(deny_severity, ALL_service.deny_severity, current_service.deny_severity);
+	userresolvreq = SET_VAR(userresolvreq, ALL_service.ident, current_service.ident);
+	defaultallow = SET_VAR(defaultallow, ALL_service.defaultrule, current_service.defaultrule);
+	ignore_src_dns_fake =SET_VAR(ignore_src_dns_fake, ALL_service.ignore_src_dns_fake, current_service.ignore_src_dns_fake);
+	ignore_dst_dns_fake = SET_VAR(ignore_dst_dns_fake, ALL_service.ignore_dst_dns_fake, current_service.ignore_dst_dns_fake);
+
+	if (ret == TCPD_ALLOW && !ignore_src_dns_fake && req->src_wrongdns)
+		ret = TCPD_DENY;
+	else
+	if (ret == TCPD_ALLOW && !ignore_dst_dns_fake && req->dst_wrongdns)
 		ret = TCPD_DENY;
 	else
 	if (ret == TCPD_ERROR) {
