@@ -14,7 +14,7 @@
 #endif
 
 #include "rcsid.h"
-RCSID("$Id: tcpd_tools.c,v 1.5 2000/05/15 09:26:13 misiek Exp $")
+RCSID("$Id: tcpd_tools.c,v 1.7 2000/10/01 21:19:57 baggins Exp $")
 
 #include <stdlib.h>
 #include <unistd.h>
@@ -272,6 +272,18 @@ int tcpd_parse_options(char *str, char **left, struct _code *flags)
 	        return TCPD_ERROR;
 	      }
 	      break;
+            case T_FUNC_SET0:
+	      if (((int (*)(int))flags[i].ptr)(0) == 0) {
+		*left = str;
+	        return TCPD_ERROR;
+	      }
+	      break;
+            case T_FUNC_SET1:
+	      if (((int (*)(int))flags[i].ptr)(1) == 0) {
+		*left = str;
+	        return TCPD_ERROR;
+	      }
+	      break;
 	    default:
 	      *left = str;
 	      return TCPD_ERROR;
@@ -299,3 +311,17 @@ int tcpd_parse_options(char *str, char **left, struct _code *flags)
 
   return TCPD_NOERROR;
 }
+
+char *tcpd_lineno(struct request_info *req)
+{
+	static char buff[TCPD_STRINGLEN];
+
+	if (req->lineno > 0)
+		snprintf(buff, sizeof(buff), "at line %d", req->lineno);
+	else if (req->lineno == -1)
+		snprintf(buff, sizeof(buff), "by default configuration");
+	else
+		snprintf(buff, sizeof(buff), "at unknown line (%d?)", req->lineno);
+	return (buff);
+}
+
